import asyncio
import logging

import attr

from uecp.commands import (
    DataSetSelectCommand,
    MessageAcknowledgementCommand,
    RDSEnabledSetCommand,
    RequestCommand,
    ResponseCode,
    UECPCommand,
)
from uecp.frame import UECPFrame
from uecp.serial_con.protocol import open_serial_protocol, UECPSerialProtocol


@attr.s(
    auto_detect=True,
    repr=True,
    order=False,
    hash=False,
    kw_only=True,
)
class GenericRDSEncoderState:
    active_data_set: int = attr.ib(
        converter=int, validator=attr.validators.in_(range(1, 9))
    )
    rds_enabled: bool = attr.ib(default=True, converter=bool)

    def __attrs_post_init__(self):
        self.logger = logging.getLogger(self.__class__.__qualname__)

    @classmethod
    async def init_from_device(
        cls, proto: UECPSerialProtocol
    ) -> "GenericRDSEncoderState":
        logger = logging.getLogger(cls.__qualname__)

        event_connection_made = asyncio.Event()
        event_all_data_received = asyncio.Event()
        proto.connection_made_callbacks.append(event_connection_made.set)

        state_dict = {}

        def frame_callback(received_frame: UECPFrame):
            for cmd in received_frame.commands:
                if isinstance(cmd, MessageAcknowledgementCommand):
                    if cmd.code is ResponseCode.OK:
                        logger.info("Last frame / cmd successfully transmitted")
                    else:
                        raise Exception(
                            "Error received as response from rds encoder", cmd
                        )
                elif isinstance(cmd, DataSetSelectCommand):
                    logger.info(f"DataSetSelectCommand received {cmd}")
                    state_dict["active_data_set"] = cmd.select_data_set_number
                elif isinstance(cmd, RDSEnabledSetCommand):
                    logger.info(f"RDSEnabledSetCommand received {cmd}")
                    state_dict["rds_enabled"] = cmd.enable
                else:
                    logger.warning("Unknown cmd received {cmd}")

            if state_dict.keys() == attr.fields_dict(cls).keys():
                event_all_data_received.set()

        proto.received_frame_callbacks.append(frame_callback)

        frame = UECPFrame(
            commands=[
                RequestCommand(command=DataSetSelectCommand),
                RequestCommand(command=RDSEnabledSetCommand),
            ]
        )
        if not proto.connected:
            await event_connection_made.wait()
        proto.connection_made_callbacks.remove(event_connection_made.set)
        proto.write(frame)
        await event_all_data_received.wait()
        proto.received_frame_callbacks.remove(frame_callback)
        current = GenericRDSEncoderState(**state_dict)

        return current

    @property
    def refresh_frames(self) -> list[UECPFrame]:
        frame = UECPFrame(
            commands=[
                RequestCommand(command=DataSetSelectCommand),
                RequestCommand(command=RDSEnabledSetCommand),
            ]
        )

        return [frame]

    def receive_frame_callback(self, received_frame: UECPFrame):
        for cmd in received_frame.commands:
            if isinstance(cmd, MessageAcknowledgementCommand):
                if cmd.code is ResponseCode.OK:
                    self.logger.info("Last frame / cmd successfully transmitted")
                else:
                    raise Exception("Error received as response from rds encoder", cmd)
            elif isinstance(cmd, DataSetSelectCommand):
                self.logger.info(f"DataSetSelectCommand received {cmd}")
                self.active_data_set = cmd.select_data_set_number
            elif isinstance(cmd, RDSEnabledSetCommand):
                self.logger.info(f"RDSEnabledSetCommand received {cmd}")
                self.rds_enabled = cmd.enable
            else:
                self.logger.warning("Unknown cmd received {cmd}")

    @staticmethod
    def compare_and_generate(
        current: "GenericRDSEncoderState", target: "GenericRDSEncoderState"
    ) -> list[UECPFrame]:
        cmds: list[UECPCommand] = []
        if current == target:
            return []

        if current.active_data_set != target.active_data_set:
            cmds.append(
                DataSetSelectCommand(select_data_set_number=target.active_data_set)
            )
        if current.rds_enabled != target.rds_enabled:
            cmds.append(RDSEnabledSetCommand(enable=target.rds_enabled))

        frames: list[UECPFrame] = []
        current_frame = UECPFrame()
        for cmd in cmds:
            try:
                current_frame.add_command(cmd)
            except OverflowError:
                frames.append(current_frame)
                current_frame = UECPFrame()
                current_frame.add_command(cmd)

        if len(current_frame.commands) > 0:
            frames.append(current_frame)

        return frames


class GenericRDSEncoder:
    def __init__(self, protocol: UECPSerialProtocol, current: GenericRDSEncoderState):
        self.logger = logging.getLogger(self.__class__.__qualname__)
        self._active_data_set: int = 0
        self._protocol: UECPSerialProtocol = protocol

        self._current = current
        self._target = attr.evolve(current)
        self._protocol.received_frame_callbacks.append(
            self._current.receive_frame_callback
        )

    @classmethod
    async def create(cls, port: str, baudrate: int) -> "GenericRDSEncoder":
        proto = await open_serial_protocol(port, baudrate)
        current = await GenericRDSEncoderState.init_from_device(proto)

        self = cls(proto, current)

        return self

    @property
    def state(self) -> GenericRDSEncoderState:
        return self._target

    @state.setter
    def state(self, value: GenericRDSEncoderState):
        self._target = value

    def ensure_current(self):
        frames = GenericRDSEncoderState.compare_and_generate(
            self._current, self._target
        )
        if len(frames) > 0:
            for frame in frames:
                self._protocol.write(frame)



import struct
import typing
import zoneinfo
from datetime import datetime, timedelta, timezone

from uecp.commands import UECPCommand
from uecp.commands.base import (
    UECPCommandDecodeElementCodeMismatchError,
    UECPCommandDecodeNotEnoughData,
)


@UECPCommand.register_type
class RealTimeClockSetCommand(UECPCommand):
    ELEMENT_CODE = 0x0D

    UTC = zoneinfo.ZoneInfo("UTC")

    def __init__(self, timestamp: typing.Optional[datetime] = None):
        self._timestamp = datetime.now(self.UTC)
        if timestamp is not None:
            self.timestamp = timestamp

    @property
    def timestamp(self) -> datetime:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: typing.Optional[datetime]):
        if value is None:
            self._timestamp = datetime.now(self.UTC)
            return
        if value.tzinfo is None:
            raise ValueError("Missing timezone")
        self._timestamp = value

    def encode(self) -> list[int]:
        ts = self._timestamp.astimezone(self.UTC)
        data = [
            self.ELEMENT_CODE,
            ts.year % 100,
            ts.month,
            ts.day,
            ts.hour,
            ts.minute,
            ts.second,
            round(ts.microsecond / 10_000),
            self._encode_localtime_offset(self._timestamp.utcoffset() or timedelta()),
        ]

        return data

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RealTimeClockSetCommand", int]:
        data = list(data)
        if len(data) < 9:
            raise UECPCommandDecodeNotEnoughData(len(data), 9)

        (
            mec,
            year,
            month,
            day,
            hour,
            minute,
            second,
            centisecond,
            encoded_localtime_offset,
        ) = data

        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        offset = cls._decode_localtime_offset(encoded_localtime_offset)
        timestamp = datetime(
            year=year + 2000,
            month=month,
            day=day,
            hour=hour,
            minute=minute,
            second=second,
            microsecond=centisecond * 10_000,
            tzinfo=cls.UTC,
        ).astimezone(timezone(offset))

        return cls(timestamp=timestamp), 9

    @staticmethod
    def _encode_localtime_offset(offset: timedelta):
        offset_seconds = offset.total_seconds()
        sign = offset_seconds < 0

        localtime_offset = (sign << 6) | abs(round(offset_seconds / 1800))

        return localtime_offset

    @staticmethod
    def _decode_localtime_offset(encoded_localtime_offset: int) -> timedelta:
        if not (0x00 <= encoded_localtime_offset <= 0x3F):
            raise ValueError(f"{encoded_localtime_offset:x} not in [0x00, 0x3f]")

        sign = -1 if encoded_localtime_offset & (1 << 6) else 1
        offset_seconds = sign * (encoded_localtime_offset & 0b11111) * 1800

        return timedelta(seconds=offset_seconds)


@UECPCommand.register_type
class RealTimeClockCorrectionSetCommand(UECPCommand):
    ELEMENT_CODE = 0x09

    _STRUCT = struct.Struct(">h")

    def __init__(self, adjustment_ms: int = 0):
        self._adjustment_ms = 0
        self.adjustment_ms = adjustment_ms

    @property
    def adjustment_ms(self) -> int:
        return self._adjustment_ms

    @adjustment_ms.setter
    def adjustment_ms(self, value):
        if not (-32768 <= value <= 32767):
            raise ValueError()
        self._adjustment_ms = int(value)

    def encode(self) -> list[int]:
        return [self.ELEMENT_CODE] + list(self._STRUCT.pack(self._adjustment_ms))

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RealTimeClockCorrectionSetCommand", int]:
        data = list(data)
        if len(data) < 3:
            raise UECPCommandDecodeNotEnoughData(len(data), 3)
        mec = data[0]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        adjustment_ms = cls._STRUCT.unpack(bytes(data[1:3]))[0]

        return cls(adjustment_ms=adjustment_ms), 3


@UECPCommand.register_type
class RealTimeClockEnabledSetCommand(UECPCommand):
    ELEMENT_CODE = 0x19

    def __init__(self, enable: bool):
        self._enable = bool(enable)

    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, value):
        self._enable = bool(value)

    def encode(self) -> list[int]:
        return [self.ELEMENT_CODE, int(self._enable)]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RealTimeClockEnabledSetCommand", int]:
        data = list(data)
        if len(data) < 2:
            raise UECPCommandDecodeNotEnoughData(len(data), 2)
        mec, enable = data[0:2]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        if enable not in (0x00, 0x01):
            raise ValueError("Not allowed value decoded")
        return cls(enable=bool(enable)), 2



import codecs
import typing

encoding_base_table: dict[int, int] = {i: i for i in range(0x20, 0x7F)}
encoding_base_table.update(
    {
        # the only control chars allowed
        ord("\r"): ord("\r"),
        ord("\n"): ord("\n"),
        # chars not matching ascii
        0x24: 0x00A4,
        0x5E: 0x2015,
        0x60: 0x2551,
        0x7E: 0x00AF,
        # custom character mapping
        0x80: 0x00E1,
        0x81: 0x00E0,
        0x82: 0x00E9,
        0x83: 0x00E8,
        0x84: 0x00ED,
        0x85: 0x00EC,
        0x86: 0x00F3,
        0x87: 0x00F2,
        0x88: 0x00FA,
        0x89: 0x00F9,
        0x8A: 0x00D1,
        0x8B: 0x00C7,
        0x8C: 0x015E,
        0x8D: 0x00DF,
        0x8E: 0x00A1,
        0x8F: 0x0132,
        0x90: 0x00E2,
        0x91: 0x00E4,
        0x92: 0x00EA,
        0x93: 0x00EB,
        0x94: 0x00EE,
        0x95: 0x00EF,
        0x96: 0x00F4,
        0x97: 0x00F6,
        0x98: 0x00FB,
        0x99: 0x00FC,
        0x9A: 0x00F1,
        0x9B: 0x00E7,
        0x9C: 0x015F,
        0x9D: 0x011F,
        0x9E: 0x0131,
        0x9F: 0x0133,
        0xA0: 0x00AA,
        0xA1: 0x03B1,
        0xA2: 0x00A9,
        0xA3: 0x2030,
        0xA4: 0x011E,
        0xA5: 0x011B,
        0xA6: 0x0148,
        0xA7: 0x0151,
        0xA8: 0x03C0,
        0xA9: 0x20AC,
        0xAA: 0x00A3,
        0xAB: 0x0024,
        0xAC: 0x2190,
        0xAD: 0x2191,
        0xAE: 0x2192,
        0xAF: 0x2193,
        0xB0: 0x00BA,
        0xB1: 0x00B9,
        0xB2: 0x00B2,
        0xB3: 0x00B3,
        0xB4: 0x00B1,
        0xB5: 0x0130,
        0xB6: 0x0144,
        0xB7: 0x0171,
        0xB8: 0x00B5,
        0xB9: 0x00BF,
        0xBA: 0x00F7,
        0xBB: 0x00B0,
        0xBC: 0x00BC,
        0xBD: 0x00BD,
        0xBE: 0x00BE,
        0xBF: 0x00A7,
        0xC0: 0x00C1,
        0xC1: 0x00C0,
        0xC2: 0x00C9,
        0xC3: 0x00C8,
        0xC4: 0x00CD,
        0xC5: 0x00CC,
        0xC6: 0x00D3,
        0xC7: 0x00D2,
        0xC8: 0x00DA,
        0xC9: 0x00D9,
        0xCA: 0x0158,
        0xCB: 0x010C,
        0xCC: 0x0160,
        0xCD: 0x017D,
        0xCE: 0x00D0,
        0xCF: 0x013F,
        0xD0: 0x00C2,
        0xD1: 0x00C4,
        0xD2: 0x00CA,
        0xD3: 0x00CB,
        0xD4: 0x00CE,
        0xD5: 0x00CF,
        0xD6: 0x00D4,
        0xD7: 0x00D6,
        0xD8: 0x00DB,
        0xD9: 0x00DC,
        0xDA: 0x0159,
        0xDB: 0x010D,
        0xDC: 0x0161,
        0xDD: 0x017E,
        0xDF: 0x0140,
        0xE0: 0x00C3,
        0xE1: 0x00C5,
        0xE2: 0x00C6,
        0xE3: 0x0152,
        0xE4: 0x0177,
        0xE5: 0x00DD,
        0xE6: 0x00D5,
        0xE7: 0x00D8,
        0xE8: 0x00DE,
        0xE9: 0x014A,
        0xEA: 0x0154,
        0xEB: 0x0106,
        0xEC: 0x015A,
        0xED: 0x0179,
        0xEE: 0x0166,
        0xEF: 0x00F0,
        0xF0: 0x00E3,
        0xF1: 0x00E5,
        0xF2: 0x00E6,
        0xF3: 0x0153,
        0xF4: 0x0175,
        0xF5: 0x00FD,
        0xF6: 0x00F5,
        0xF7: 0x00F8,
        0xF8: 0x00FE,
        0xF9: 0x014B,
        0xFA: 0x0155,
        0xFB: 0x0107,
        0xFC: 0x015B,
        0xFD: 0x017A,
        0xFE: 0x0167,
    }
)

encoding_table: dict[str, int] = {}
decoding_table: dict[int, str] = {}

for (character_code, unicode_number) in encoding_base_table.items():
    unicode_char = chr(unicode_number)

    if unicode_char in encoding_table:
        raise Exception(
            f"unicode_char {unicode_char} already mapped to {encoding_table[unicode_char]:#x}"
        )
    encoding_table[unicode_char] = character_code

    if character_code in decoding_table:
        raise Exception()
    decoding_table[character_code] = unicode_char


def encode(input_string: str, errors: str = "strict") -> tuple[bytes, int]:
    encoded = []
    encoded_chars = 0

    for i, char in enumerate(input_string):
        try:
            encoded.append(encoding_table[char])
            encoded_chars += 1
        except KeyError:
            if errors == "strict":
                raise UnicodeError(
                    f"Cannot encode {char}, code point {ord(char):#x} at position {i}"
                )
    return bytes(encoded), encoded_chars


def decode(data: bytes, errors: str = "strict") -> tuple[str, int]:
    decoded = ""
    decoded_chars = 0

    for i, byte in enumerate(list(data)):
        try:
            decoded += decoding_table[byte]
            decoded_chars += 1
        except KeyError:
            if errors == "strict":
                raise UnicodeError(f"Cannot decode byte {byte:#x} at position {i}")

    return decoded, decoded_chars


class Codec(codecs.Codec):
    def encode(self, input_string: str, errors: str = "strict") -> tuple[bytes, int]:
        return encode(input_string, errors)

    def decode(self, data: bytes, errors: str = "strict") -> tuple[str, int]:
        return decode(data, errors)


class IncrementalEncoder(codecs.IncrementalEncoder):
    def encode(self, data: str, final: bool = False) -> bytes:
        return encode(data, errors=self.errors)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):
    def decode(self, data: bytes, final: bool = False) -> str:
        return decode(data, errors=self.errors)[0]


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass


def getregentry():
    """encodings module API"""
    return codecs.CodecInfo(
        name="basic_rds_character_set",
        encode=encode,
        decode=decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
        _is_text_encoding=True,
    )


def search_function(name: str) -> typing.Optional[codecs.CodecInfo]:
    if name in ["basic_rds_character_set", "rds_character_set", "rds"]:
        return getregentry()
    return None


codecs.register(search_function)



from uecp.commands.base import UECPCommandException


class InvalidDataSetNumber(UECPCommandException):
    def __init__(self, data_set_number):
        self.data_set_number = data_set_number

    def __repr__(self) -> str:
        return f"InvalidDataSetNumber(data_set_number={self.data_set_number!r})"

    def __str__(self):
        return (
            f"Supplied an invalid value for data set number. "
            f"Must be an integer between 0x00 and 0xFF. Supplied {self.data_set_number!r}"
        )


class UECPCommandDataSetNumber:
    CURRENT_DATA_SET = 0x00
    ALL_EXCEPT_CURRENT_DATA_SET = 0xFE
    ALL_DATA_SETS = 0xFF

    def __init__(self, data_set_number=0, **kwargs):
        super().__init__(**kwargs)
        self.__data_set_number = 0

        self.data_set_number = data_set_number

    @property
    def data_set_number(self) -> int:
        return self.__data_set_number

    @data_set_number.setter
    def data_set_number(self, new_data_set_number: int):
        try:
            if new_data_set_number == int(new_data_set_number):
                new_data_set_number = int(new_data_set_number)
            else:
                raise ValueError()
        except ValueError:
            raise InvalidDataSetNumber(new_data_set_number)

        if not (0x00 <= new_data_set_number <= 0xFF):
            raise InvalidDataSetNumber(new_data_set_number)
        self.__data_set_number = new_data_set_number


class InvalidProgrammeServiceNumber(UECPCommandException):
    def __init__(self, programme_service_number):
        self.programme_service_number = programme_service_number

    def __repr__(self) -> str:
        return f"InvalidProgrammeServiceNumber(programme_service_number={self.programme_service_number!r})"

    def __str__(self):
        return (
            f"Supplied an invalid value for programme service number. "
            f"Must be an integer between 0x00 and 0xFF. Supplied {self.programme_service_number!r}"
        )


class UECPCommandProgrammeServiceNumber:
    def __init__(self, programme_service_number=0, **kwargs):
        super().__init__(**kwargs)
        self.__programme_service_number = 0

        self.programme_service_number = programme_service_number

    @property
    def programme_service_number(self) -> int:
        return self.__programme_service_number

    @programme_service_number.setter
    def programme_service_number(self, new_programme_service_number: int):
        try:
            if new_programme_service_number == int(new_programme_service_number):
                new_programme_service_number = int(new_programme_service_number)
            else:
                raise ValueError
        except ValueError:
            raise InvalidProgrammeServiceNumber(new_programme_service_number)

        if not (0x00 <= new_programme_service_number <= 0xFF):
            raise InvalidProgrammeServiceNumber(new_programme_service_number)
        self.__programme_service_number = new_programme_service_number


class UECPCommandDSNnPSN(UECPCommandDataSetNumber, UECPCommandProgrammeServiceNumber):
    pass



import enum
import typing

from uecp.commands import UECPCommand
from uecp.commands.base import (
    UECPCommandDecodeElementCodeMismatchError,
    UECPCommandDecodeNotEnoughData,
)


@enum.unique
class ResponseCode(enum.IntEnum):
    OK = 0
    CRC_ERROR = 1
    MSG_NOT_RECEIVED = 2
    MSG_UNKNOWN = 3
    DSN_ERROR = 4
    PSN_ERROR = 5
    PARAM_OUT_OF_RANGE = 6
    MSG_ELEMENT_LENGTH_ERROR = 7
    MSG_FIELD_LENGTH_ERROR = 8
    MSG_NOT_ACCEPTABLE = 9
    END_MSG_MISSING = 10
    BUFFER_OVERFLOW = 11
    BAD_STUFFING = 12
    UNEXPECTED_END_OF_MSG = 13


@UECPCommand.register_type
class MessageAcknowledgementCommand(UECPCommand):
    ELEMENT_CODE = 0x18

    def __init__(self, code: ResponseCode, sequence_counter: int = 0):
        self._code = ResponseCode(code)
        self._sequence_counter = 0
        self.sequence_counter = sequence_counter

    @property
    def code(self) -> ResponseCode:
        return self._code

    @code.setter
    def code(self, value):
        self._code = ResponseCode(value)

    @property
    def sequence_counter(self) -> int:
        return self._sequence_counter

    @sequence_counter.setter
    def sequence_counter(self, new_sequence_counter: int):
        if not (0 <= new_sequence_counter <= 0xFF):
            raise ValueError(
                f"Sequence counter must be in range of 0 to 0xff, {new_sequence_counter:#x} given"
            )
        if new_sequence_counter != int(new_sequence_counter):
            raise ValueError("Sequence counter must be an integer")
        self._sequence_counter = int(new_sequence_counter)

    def encode(self) -> list[int]:
        if self._code is not ResponseCode.OK:
            return [self.ELEMENT_CODE, int(self._code), self._sequence_counter]
        else:
            return [self.ELEMENT_CODE, 0]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["MessageAcknowledgementCommand", int]:
        data = list(data)
        if len(data) < 2:
            raise UECPCommandDecodeNotEnoughData(len(data), 2)
        mec, code = data[0:2]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        code = ResponseCode(code)
        if code is ResponseCode.OK:
            return cls(code=code), 2
        if len(data) < 3:
            raise UECPCommandDecodeNotEnoughData(len(data), 3)
        sequence_counter = data[2]
        return cls(code=code, sequence_counter=sequence_counter), 3


@UECPCommand.register_type
class RequestCommand(UECPCommand):
    ELEMENT_CODE = 0x17

    def __init__(
        self,
        *,
        element_code: typing.Optional[int] = None,
        command: typing.Union[UECPCommand, type[UECPCommand]] = None,
        data_set_number: typing.Optional[int] = None,
        programme_service_number: typing.Optional[int] = None,
        additional_data: typing.Union[bytes, list[int]] = None,
    ):
        if element_code is not None:
            if command is not None:
                raise ValueError(
                    "element_code and command are mutually exclusive required"
                )
            if element_code not in UECPCommand.ELEMENT_CODE_MAP:
                raise ValueError(f"Unknown element code {element_code:#x}")
            command = UECPCommand.ELEMENT_CODE_MAP[element_code]
        if command is None:
            raise ValueError("element_code or command are mutually exclusive required")
        if hasattr(command, "data_set_number") and data_set_number is None:
            raise ValueError("command requires data set number")
        if (
            hasattr(command, "programme_service_number")
            and programme_service_number is None
        ):
            raise ValueError("command requires programme service number")

        self._element_code = command.ELEMENT_CODE
        self._dsn = data_set_number
        self._psn = programme_service_number
        self._additional_data = (
            list(additional_data) if additional_data is not None else []
        )

    @property
    def element_code(self) -> int:
        return self._element_code

    @property
    def data_set_number(self) -> typing.Optional[int]:
        return self._dsn

    @property
    def programme_service_number(self) -> typing.Optional[int]:
        return self._psn

    def encode(self) -> list[int]:
        request_data = [self._element_code]
        if self._dsn is not None:
            request_data.append(self._dsn)
        if self._psn is not None:
            request_data.append(self._psn)
        if self._additional_data:
            request_data += self._additional_data

        return [self.ELEMENT_CODE, len(request_data)] + request_data

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RequestCommand", int]:
        data = list(data)
        if len(data) < 2:
            raise UECPCommandDecodeNotEnoughData(len(data), 2)
        mec, mel = data[0:2]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        if len(data) < (2 + mel):
            raise UECPCommandDecodeNotEnoughData(len(data), 2 + mel)
        data, idx = data[2:], 0
        element_code, idx = data[idx], idx + 1
        if element_code not in UECPCommand.ELEMENT_CODE_MAP:
            raise ValueError(f"Unknown element code {element_code:#x}")
        command = UECPCommand.ELEMENT_CODE_MAP[element_code]
        data_set_number = None
        if hasattr(command, "data_set_number"):
            data_set_number, idx = data[idx], idx + 1
        programme_service_number = None
        if hasattr(command, "programme_service_number"):
            programme_service_number, idx = data[idx], idx + 1
        additional_data = data[idx:mel]
        return (
            cls(
                command=command,
                data_set_number=data_set_number,
                programme_service_number=programme_service_number,
                additional_data=additional_data,
            ),
            2 + mel,
        )



import abc
import typing

T_UECPCommand = typing.TypeVar("T_UECPCommand", bound="UECPCommand")


class UECPCommand(abc.ABC):
    ELEMENT_CODE: typing.ClassVar[int]
    ELEMENT_CODE_MAP: typing.ClassVar[dict[int, type["UECPCommand"]]] = {}

    @abc.abstractmethod
    def encode(self) -> list[int]:
        ...

    @classmethod
    @abc.abstractmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["UECPCommand", int]:
        ...

    @classmethod
    def register_type(cls, message_type: type[T_UECPCommand]) -> type[T_UECPCommand]:
        mec = int(message_type.ELEMENT_CODE)
        if not (0x01 <= mec <= 0xFD):
            raise ValueError(f"MEC must be in [0x01, 0xFD], given {mec:#x}")
        if mec in cls.ELEMENT_CODE_MAP:
            raise ValueError(f"MEC {mec:#x} already defined")
        cls.ELEMENT_CODE_MAP[mec] = message_type
        return message_type

    @classmethod
    def decode_commands(
        cls, data: typing.Union[bytes, list[int]]
    ) -> list["UECPCommand"]:
        cmds = []
        data = list(data)
        while len(data) > 0:
            mec = data[0]
            if mec not in cls.ELEMENT_CODE_MAP:
                raise ValueError()
            cmd, consumed_bytes = cls.ELEMENT_CODE_MAP[mec].create_from(data)
            cmds.append(cmd)
            data = data[consumed_bytes:]
        return cmds


class UECPCommandException(Exception):
    pass


class UECPCommandDecodeError(UECPCommandException):
    pass


class UECPCommandDecodeNotEnoughData(UECPCommandDecodeError):
    pass


class UECPCommandDecodeElementCodeMismatchError(UECPCommandDecodeError):
    pass



import typing

from uecp.commands.base import (
    UECPCommand,
    UECPCommandDecodeElementCodeMismatchError,
    UECPCommandDecodeNotEnoughData,
)


@UECPCommand.register_type
class RDSEnabledSetCommand(UECPCommand):
    ELEMENT_CODE = 0x1E

    def __init__(self, enable: bool):
        self._enable = bool(enable)

    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, value):
        self._enable = bool(value)

    def encode(self) -> list[int]:
        return [self.ELEMENT_CODE, int(self._enable)]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RDSEnabledSetCommand", int]:
        data = list(data)
        if len(data) < 2:
            raise UECPCommandDecodeNotEnoughData(len(data), 2)
        mec, enable = data[0:2]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        if enable not in (0x00, 0x01):
            raise ValueError("Not allowed value decoded")
        return cls(enable=bool(enable)), 2


@UECPCommand.register_type
class RDSPhaseSetCommand(UECPCommand):
    ELEMENT_CODE = 0x22

    ALL_REFERENCE_TABLES = 0
    CURRENT_REFERENCE_TABLE = 7

    def __init__(self, reference_table: int, deci_degrees: int):
        self._reference_table = 0
        self.reference_table = reference_table
        self._deci_degrees = 0
        self.deci_degrees = deci_degrees

    @property
    def reference_table(self):
        return self._reference_table

    @reference_table.setter
    def reference_table(self, value):
        if not (0 <= value <= 7):
            raise ValueError()
        self._reference_table = value

    @property
    def deci_degrees(self):
        return self._deci_degrees

    @deci_degrees.setter
    def deci_degrees(self, value):
        if not (0 <= value <= 3599):
            raise ValueError()
        self._deci_degrees = int(value)

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            ((self._reference_table << 5) | (self._deci_degrees >> 8)),
            self._deci_degrees & 0xFF,
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RDSPhaseSetCommand", int]:
        data = list(data)
        if len(data) < 3:
            raise UECPCommandDecodeNotEnoughData(len(data), 3)
        mec = data[0]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        reference_table = (data[1] & (0b111 << 5)) >> 5
        deci_degrees = (data[1] & 0b1111) << 8 | data[2]

        return cls(reference_table=reference_table, deci_degrees=deci_degrees), 3


@UECPCommand.register_type
class RDSLevelSetCommand(UECPCommand):
    ELEMENT_CODE = 0x0E

    def __init__(self, reference_table: int, level: int):
        self._reference_table = 0
        self.reference_table = reference_table
        self._level = 0
        self.level = level

    @property
    def reference_table(self):
        return self._reference_table

    @reference_table.setter
    def reference_table(self, value):
        if not (0 <= value <= 7):
            raise ValueError()
        self._reference_table = value

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, value):
        if not (0 <= value <= 8191):
            raise ValueError()
        self._level = int(value)

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            (self.reference_table << 5) | (self._level >> 8),
            self._level & 0xFF,
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RDSLevelSetCommand", int]:
        data = list(data)
        if len(data) < 3:
            raise UECPCommandDecodeNotEnoughData(len(data), 3)
        mec = data[0]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        reference_table = (data[1] & 0b1110_0000) >> 5
        level = (data[1] & 0b11111) << 8 | data[2]

        return cls(reference_table=reference_table, level=level), 3



from uecp.commands.base import UECPCommand
from uecp.commands.bidirectional import (
    MessageAcknowledgementCommand,
    RequestCommand,
    ResponseCode,
)
from uecp.commands.clock_control import (
    RealTimeClockCorrectionSetCommand,
    RealTimeClockEnabledSetCommand,
    RealTimeClockSetCommand,
)
from uecp.commands.control_n_setup import (
    CommunicationMode,
    CommunicationModeSetCommand,
    DataSetSelectCommand,
    EncoderAddressSetCommand,
    SiteAddressSetCommand,
    SiteEncoderAddressSetCommandMode,
)
from uecp.commands.mixins import InvalidDataSetNumber, InvalidProgrammeServiceNumber
from uecp.commands.rds_control import (
    RDSEnabledSetCommand,
    RDSLevelSetCommand,
    RDSPhaseSetCommand,
)
from uecp.commands.rds_message import (
    DecoderInformationSetCommand,
    InvalidNumberOfTransmissions,
    InvalidProgrammeIdentification,
    InvalidProgrammeServiceName,
    InvalidProgrammeTypeName,
    ProgrammeIdentificationSetCommand,
    ProgrammeServiceNameSetCommand,
    ProgrammeType,
    ProgrammeTypeNameSetCommand,
    ProgrammeTypeSetCommand,
    RadioText,
    RadioTextBufferConfiguration,
    RadioTextSetCommand,
    TrafficAnnouncementProgrammeSetCommand,
)



import asyncio
import logging
import typing
from asyncio import transports
from typing import Optional

import serial  # type: ignore
import serial_asyncio  # type: ignore

from uecp.frame import UECPFrame, UECPFrameDecoder


class UECPSerialProtocol(asyncio.Protocol):
    def __init__(self) -> None:
        self.logger = logging.getLogger(self.__class__.__qualname__)
        self._transport: Optional[serial_asyncio.SerialTransport] = None

        self._uecp_frame_decoder = UECPFrameDecoder()

        self.connection_made_callbacks: list[typing.Callable[[], None]] = []
        self.received_frame_callbacks: list[typing.Callable[[UECPFrame], None]] = []

    @property
    def connected(self) -> bool:
        if self._transport:
            return not self._transport.is_closing()
        return False

    def connection_made(self, transport: transports.BaseTransport):
        self.logger.debug(f"Connection made {transport}")
        if self._transport is not None:
            raise ValueError("Connection already open?")
        self._transport = transport
        for callback in self.connection_made_callbacks:
            callback()

    def connection_lost(self, exc: Optional[Exception]):
        self._transport = None
        if exc is None:
            if not self._uecp_frame_decoder.empty:
                raise Exception("Interrupted within decoding a frame")
            return
        raise exc

    def data_received(self, data: bytes):
        self.logger.debug(f"Data received {data.hex()}")

        frame, remaining_data = self._uecp_frame_decoder.decode(data)
        while frame is not None:
            if frame:
                for callback in self.received_frame_callbacks:
                    callback(frame)
            frame, remaining_data = self._uecp_frame_decoder.decode(remaining_data)
        if len(remaining_data) > 0:
            raise Exception(
                f"not all received bytes could be decoded, should never happen, {remaining_data!r}"
            )

    def write(self, frame: UECPFrame):
        if self._transport:
            data = frame.encode()
            self.logger.debug(f"Writing {data.hex()}")
            self._transport.write(data)
        else:
            self.logger.error("No transport opened yet")


async def open_serial_protocol(port: str, baudrate: int) -> UECPSerialProtocol:
    con = serial.Serial(
        port=port,
        baudrate=baudrate,
        bytesize=serial.EIGHTBITS,
        parity=serial.PARITY_NONE,
        stopbits=serial.STOPBITS_ONE,
        xonxoff=False,
        rtscts=False,
        dsrdtr=False,
    )

    _, protocol = await serial_asyncio.connection_for_serial(
        asyncio.get_running_loop(), UECPSerialProtocol, con
    )
    return protocol



import codecs
import typing


def encode(
    data: typing.Union[list[int], bytes], errors: str = "strict"
) -> tuple[bytes, int]:
    stuffed_data = []
    for byte in list(data):
        if byte == 0xFD:
            stuffed_data += [0xFD, 0x00]
        elif byte == 0xFE:
            stuffed_data += [0xFD, 0x01]
        elif byte == 0xFF:
            stuffed_data += [0xFD, 0x02]
        else:
            stuffed_data += [byte]

    return bytes(stuffed_data), len(data)


def decode(
    stuffed_data: typing.Union[list[int], bytes], errors: str = "strict"
) -> tuple[bytes, int]:
    data = []
    next_byte_stuffed = False
    decoded_bytes = len(stuffed_data)
    for i, byte in enumerate(list(stuffed_data)):
        if 0x00 <= byte < 0xFD and not next_byte_stuffed:
            data.append(byte)
        elif byte == 0xFD and not next_byte_stuffed:
            next_byte_stuffed = True
        elif 0x00 <= byte <= 0x02 and next_byte_stuffed:
            data.append(byte + 0xFD)
            next_byte_stuffed = False
        else:
            if errors == "strict":
                raise UnicodeError(
                    f"Unknown state, byte {byte:#x}, next_byte_stuffed {next_byte_stuffed}, at position {i}"
                )
            decoded_bytes -= 1

    if next_byte_stuffed and errors == "strict":
        raise UnicodeError(f"Expecting one more byte as byte stuffing must be happened")

    return bytes(data), decoded_bytes


class Codec(codecs.Codec):
    def encode(self, data: bytes, errors: str = "strict") -> tuple[bytes, int]:  # type: ignore[override]
        return encode(data)

    def decode(self, data: bytes, errors: str = "strict") -> tuple[bytes, int]:  # type: ignore[override]
        return decode(data, errors=errors)


class IncrementalEncoder(codecs.IncrementalEncoder):
    def encode(self, data: bytes, final: bool = False) -> bytes:  # type: ignore[override]
        return bytes(encode(data)[0])


class IncrementalDecoder(codecs.IncrementalDecoder):
    def __init__(self, errors: str = "strict"):
        self.errors = errors
        self.next_byte_stuffed = False

    def reset(self):
        self.next_byte_stuffed = False

    def getstate(self) -> tuple[bytes, int]:
        return bytes(), int(self.next_byte_stuffed)

    def setstate(self, state: tuple[bytes, int]):
        self.next_byte_stuffed = bool(state[1])

    def decode(  # type: ignore[override]
        self, data: typing.Union[bytes, list[int]], final: bool = False
    ) -> bytes:
        decoded = []
        for i, byte in enumerate(list(data)):
            if 0x00 <= byte < 0xFD and not self.next_byte_stuffed:
                decoded.append(byte)
            elif byte == 0xFD and not self.next_byte_stuffed:
                self.next_byte_stuffed = True
            elif 0x00 <= byte <= 0x02 and self.next_byte_stuffed:
                decoded.append(byte + 0xFD)
                self.next_byte_stuffed = False
            else:
                if self.errors == "strict":
                    raise UnicodeError(
                        f"Unknown state, byte {byte:#x}, next_byte_stuffed {self.next_byte_stuffed}, at position {i}"
                    )

        if final:
            if self.next_byte_stuffed and self.errors == "strict":
                raise UnicodeError(
                    f"Expecting one more byte as byte stuffing must be happened"
                )
            self.next_byte_stuffed = False

        return bytes(decoded)


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass


def getregentry():
    """encodings module API"""
    return codecs.CodecInfo(
        name="uecp_frame",
        encode=encode,
        decode=decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
        _is_text_encoding=False,
    )


def search_function(name: str) -> typing.Optional[codecs.CodecInfo]:
    if name in ["uecp_frame"]:
        return getregentry()
    return None


codecs.register(search_function)



import enum
import typing

from uecp.commands.base import (
    UECPCommand,
    UECPCommandDecodeElementCodeMismatchError,
    UECPCommandDecodeNotEnoughData,
)
from uecp.commands.mixins import InvalidDataSetNumber


class SiteEncoderAddressSetCommandMode(enum.IntEnum):
    REMOVE_SINGLE = 0b00
    ADD_SINGLE = 0b01
    REMOVE_ALL = 0b10


@UECPCommand.register_type
class SiteAddressSetCommand(UECPCommand):
    ELEMENT_CODE = 0x23

    def __init__(self, mode: SiteEncoderAddressSetCommandMode, site_address: int):
        self._mode: SiteEncoderAddressSetCommandMode = (
            SiteEncoderAddressSetCommandMode.ADD_SINGLE
        )
        self._site_address = 0
        self.mode = mode
        self.site_address = site_address

    @property
    def mode(self) -> SiteEncoderAddressSetCommandMode:
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = SiteEncoderAddressSetCommandMode(value)

    @property
    def site_address(self) -> int:
        return self._site_address

    @site_address.setter
    def site_address(self, value):
        value = int(value)
        if not (0 <= value <= 0x3FF):
            raise ValueError(
                f"Site address must be in range of 0 to 0x3ff, {value:#x} given"
            )
        self._site_address = value

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            int(self._mode),
            self._site_address >> 8,
            self._site_address & 0xFF,
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["SiteAddressSetCommand", int]:
        data = list(data)
        if len(data) < 4:
            raise UECPCommandDecodeNotEnoughData(len(data), 4)
        mec, mode, site_address_high, site_address_low = data[0:4]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        site_address = site_address_high << 8 | site_address_low
        self = cls(
            mode=SiteEncoderAddressSetCommandMode(mode), site_address=site_address
        )
        return self, 4


@UECPCommand.register_type
class EncoderAddressSetCommand(UECPCommand):
    ELEMENT_CODE = 0x27

    def __init__(self, mode: SiteEncoderAddressSetCommandMode, encoder_address: int):
        self._mode: SiteEncoderAddressSetCommandMode = (
            SiteEncoderAddressSetCommandMode.ADD_SINGLE
        )
        self._encoder_address = 0
        self.mode = mode
        self.encoder_address = encoder_address

    @property
    def mode(self) -> SiteEncoderAddressSetCommandMode:
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = SiteEncoderAddressSetCommandMode(value)

    @property
    def encoder_address(self) -> int:
        return self._encoder_address

    @encoder_address.setter
    def encoder_address(self, value):
        value = int(value)
        if not (0 <= value <= 0x3F):
            raise ValueError(
                f"Encoder address must be in range of 0 to 0x3f, {value:#x} given"
            )
        self._encoder_address = value

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            int(self._mode),
            self._encoder_address,
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["EncoderAddressSetCommand", int]:
        data = list(data)
        if len(data) < 3:
            raise UECPCommandDecodeNotEnoughData(len(data), 3)
        mec, mode, encoder_address = data[0:3]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        return (
            cls(
                mode=SiteEncoderAddressSetCommandMode(mode),
                encoder_address=encoder_address,
            ),
            3,
        )


@enum.unique
class CommunicationMode(enum.IntEnum):
    UNIDIRECTIONAL = 0
    BIDIRECTIONAL_REQUESTED_RESPONSE = 1
    BIDIRECTIONAL_SPONTANEOUS_RESPONSE = 2


@UECPCommand.register_type
class CommunicationModeSetCommand(UECPCommand):
    ELEMENT_CODE = 0x2C

    def __init__(self, mode: CommunicationMode):
        self._mode = CommunicationMode(mode)

    @property
    def mode(self) -> CommunicationMode:
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = CommunicationMode(value)

    def encode(self) -> list[int]:
        return [self.ELEMENT_CODE, int(self._mode)]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["CommunicationModeSetCommand", int]:
        data = list(data)
        if len(data) < 2:
            raise UECPCommandDecodeNotEnoughData(len(data), 2)
        mec, mode = data[0:2]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        return cls(mode=CommunicationMode(mode)), 2


@UECPCommand.register_type
class DataSetSelectCommand(UECPCommand):
    ELEMENT_CODE = 0x1C

    def __init__(self, select_data_set_number: int):
        self._select_data_set_number = 0
        self.select_data_set_number = select_data_set_number

    @property
    def select_data_set_number(self) -> int:
        return self._select_data_set_number

    @select_data_set_number.setter
    def select_data_set_number(self, value: int):
        try:
            if value == int(value):
                value = int(value)
            else:
                raise ValueError()
        except ValueError:
            raise InvalidDataSetNumber(value)

        if not (0x01 <= value <= 0xFF):
            raise InvalidDataSetNumber(value)
        self._select_data_set_number = value

    def encode(self) -> list[int]:
        return [self.ELEMENT_CODE, self._select_data_set_number]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["DataSetSelectCommand", int]:
        data = list(data)
        if len(data) < 2:
            raise UECPCommandDecodeNotEnoughData(len(data), 2)
        mec, select_data_set_number = data[0:2]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        return cls(select_data_set_number=select_data_set_number), 2



import enum
import typing
import warnings

import attr

from uecp.commands.base import (
    UECPCommand,
    UECPCommandDecodeElementCodeMismatchError,
    UECPCommandDecodeNotEnoughData,
    UECPCommandException,
)
from uecp.commands.mixins import UECPCommandDSNnPSN


# PIN 0x06 / Programme Item Number not implemented as deprecated
# MS 0x05 / Music/Speech flag deprecated


class InvalidProgrammeIdentification(UECPCommandException):
    pass


@UECPCommand.register_type
class ProgrammeIdentificationSetCommand(UECPCommand, UECPCommandDSNnPSN):
    ELEMENT_CODE = 0x01

    def __init__(self, pi=0, data_set_number=0, programme_service_number=0):
        super().__init__(
            data_set_number=data_set_number,
            programme_service_number=programme_service_number,
        )
        self.__pi = 0
        self.pi = pi

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            self.data_set_number,
            self.programme_service_number,
            self.__pi >> 8,
            self.__pi & 0xFF,
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["ProgrammeIdentificationSetCommand", int]:
        data = list(data)
        if len(data) < 5:
            raise UECPCommandDecodeNotEnoughData(len(data), 5)

        mec, dsn, psn, pi_msb, pi_lsb = data[0:5]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        pi = pi_msb << 8 | pi_lsb

        return cls(pi=pi, data_set_number=dsn, programme_service_number=psn), 5

    @property
    def pi(self) -> int:
        return self.__pi

    @pi.setter
    def pi(self, new_pi: int):
        try:
            if new_pi == int(new_pi):
                new_pi = int(new_pi)
            else:
                raise ValueError()
        except ValueError:
            raise InvalidProgrammeIdentification(new_pi)

        if not (0x0 <= new_pi <= 0xFFFF):
            raise InvalidProgrammeIdentification(new_pi)
        self.__pi = new_pi


class InvalidProgrammeServiceName(UECPCommandException):
    def __init__(self, programme_service_name, cause: str = "unknown"):
        self.programme_service_name = programme_service_name
        self.cause = str(cause)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(programme_service_name={self.programme_service_name!r}, cause={self.cause!r})"

    def __str__(self):
        return f"Supplied an invalid value for programme service name, cause: {self.cause}. Supplied {self.programme_service_name!r}"


@UECPCommand.register_type
class ProgrammeServiceNameSetCommand(UECPCommand, UECPCommandDSNnPSN):
    ELEMENT_CODE = 0x02

    def __init__(self, ps: str = "", data_set_number=0, programme_service_number=0):
        super().__init__(
            data_set_number=data_set_number,
            programme_service_number=programme_service_number,
        )
        self.__ps = ""
        self.ps = ps

    @property
    def ps(self) -> str:
        return self.__ps

    @ps.setter
    def ps(self, new_ps: str):
        new_ps = str(new_ps)
        if len(new_ps) > 8:
            raise InvalidProgrammeServiceName(new_ps, f"PS supports only 8 characters")
        new_ps = new_ps.rstrip(" ")
        try:
            new_ps.encode("basic_rds_character_set")
        except ValueError as e:
            raise InvalidProgrammeServiceName(
                new_ps, f"PS cannot be encoded, exc={e!r}"
            )
        self.__ps = new_ps

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            self.data_set_number,
            self.programme_service_number,
        ] + list(self.__ps.ljust(8).encode("basic_rds_character_set"))

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["ProgrammeServiceNameSetCommand", int]:
        data = list(data)
        if len(data) < 11:
            raise UECPCommandDecodeNotEnoughData(len(data), 11)

        mec, dsn, psn = data[0:3]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        ps = bytes(data[3:11]).decode("basic_rds_character_set")

        return cls(ps=ps, data_set_number=dsn, programme_service_number=psn), 11


@UECPCommand.register_type
class DecoderInformationSetCommand(UECPCommand, UECPCommandDSNnPSN):
    ELEMENT_CODE = 0x04

    def __init__(
        self,
        stereo=False,
        dynamic_pty=False,
        data_set_number=0,
        programme_service_number=0,
    ):
        super().__init__(
            data_set_number=data_set_number,
            programme_service_number=programme_service_number,
        )
        self.__stereo = bool(stereo)
        self.__dynamic_pty = bool(dynamic_pty)
        # artificial head & compressed deprecated; stereo / mono flagged to be checked

    @property
    def stereo(self) -> bool:
        return self.__stereo

    @stereo.setter
    def stereo(self, enabled: bool):
        self.__stereo = bool(enabled)

    @property
    def mono(self) -> bool:
        return not self.__stereo

    @mono.setter
    def mono(self, enabled: bool):
        self.__stereo = not bool(enabled)

    @property
    def dynamic_pty(self) -> bool:
        return self.__dynamic_pty

    @dynamic_pty.setter
    def dynamic_pty(self, enabled: bool):
        self.__dynamic_pty = bool(enabled)

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            self.data_set_number,
            self.programme_service_number,
            self.__dynamic_pty << 3 | self.__stereo,
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["DecoderInformationSetCommand", int]:
        data = list(data)
        if len(data) < 4:
            raise UECPCommandDecodeNotEnoughData(len(data), 4)

        mec, dsn, psn, flags = data[0:4]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        stereo = 0b1 & flags
        dynamic_pty = 0b1000 & flags

        return (
            cls(
                stereo=stereo,
                dynamic_pty=dynamic_pty,
                data_set_number=dsn,
                programme_service_number=psn,
            ),
            4,
        )


@UECPCommand.register_type
class TrafficAnnouncementProgrammeSetCommand(UECPCommand, UECPCommandDSNnPSN):
    ELEMENT_CODE = 0x03

    def __init__(
        self,
        announcement=False,
        programme=False,
        data_set_number=0,
        programme_service_number=0,
    ):
        super().__init__(
            data_set_number=data_set_number,
            programme_service_number=programme_service_number,
        )
        self.__announcement = bool(announcement)
        self.__programme = bool(programme)

    @property
    def announcement(self) -> bool:
        return self.__announcement

    @announcement.setter
    def announcement(self, enabled: bool):
        self.__announcement = bool(enabled)

    @property
    def programme(self):
        return self.__programme

    @programme.setter
    def programme(self, enabled: bool):
        self.__programme = bool(enabled)

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            self.data_set_number,
            self.programme_service_number,
            self.__programme << 1 | self.__announcement,
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["TrafficAnnouncementProgrammeSetCommand", int]:
        data = list(data)
        if len(data) < 4:
            raise UECPCommandDecodeNotEnoughData(len(data), 4)

        mec, dsn, psn, flags = data[0:4]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        announcement = 0b1 & flags
        programme = 0b10 & flags

        return (
            cls(
                announcement=announcement,
                programme=programme,
                data_set_number=dsn,
                programme_service_number=psn,
            ),
            4,
        )


@enum.unique
class ProgrammeType(enum.IntEnum):
    UNDEFINED = 0
    NEWS = 1
    CURRENT_AFFAIRS = 2
    INFORMATION = 3
    SPORT = 4
    EDUCATION = 5
    DRAMA = 6
    CULTURE = 7
    SCIENCE = 8
    VARIED = 9
    POP_MUSIC = 10
    ROCK_MUSIC = 11
    EASY_LISTENING_MUSIC = 12
    LIGHT_CLASSICAL = 13
    SERIOUS_CLASSICAL = 14
    OTHER_MUSIC = 15
    WEATHER = 16
    FINANCE = 17
    CHILDREN_PROGRAMME = 18
    SOCIAL_AFFAIRS = 19
    RELIGION = 20
    PHONE_IN = 21
    TRAVEL = 22
    LEISURE = 23
    JAZZ_MUSIC = 24
    COUNTRY_MUSIC = 25
    NATIONAL_MUSIC = 26
    OLDIES_MUSIC = 27
    FOLK_MUSIC = 28
    DOCUMENTARY = 29
    ALARM_TEST = 30
    ALARM = 31


@UECPCommand.register_type
class ProgrammeTypeSetCommand(UECPCommand, UECPCommandDSNnPSN):
    ELEMENT_CODE = 0x07

    def __init__(
        self,
        programme_type: typing.Union[ProgrammeType, int] = ProgrammeType.UNDEFINED,
        data_set_number=0,
        programme_service_number=0,
    ):
        super().__init__(
            data_set_number=data_set_number,
            programme_service_number=programme_service_number,
        )
        self.__programme_type = ProgrammeType(programme_type)

    @property
    def programme_type(self) -> ProgrammeType:
        return self.__programme_type

    @programme_type.setter
    def programme_type(self, new_programme_type: typing.Union[int, ProgrammeType]):
        self.__programme_type = ProgrammeType(new_programme_type)

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            self.data_set_number,
            self.programme_service_number,
            int(self.__programme_type),
        ]

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["ProgrammeTypeSetCommand", int]:
        data = list(data)
        if len(data) < 4:
            raise UECPCommandDecodeNotEnoughData(len(data), 4)

        mec, dsn, psn, programme_type = data[0:4]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        programme_type = ProgrammeType(programme_type)

        return (
            cls(
                programme_type=programme_type,
                data_set_number=dsn,
                programme_service_number=psn,
            ),
            4,
        )


class InvalidProgrammeTypeName(UECPCommandException):
    def __init__(self, programme_type_name, cause: str = "unknown"):
        self.programme_type_name = programme_type_name
        self.cause = str(cause)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(programme_type_name={self.programme_type_name!r}, cause={self.cause!r})"

    def __str__(self):
        return f"Supplied an invalid value for programme type name, cause: {self.cause}. Supplied {self.programme_type_name!r}"


@UECPCommand.register_type
class ProgrammeTypeNameSetCommand(UECPCommand, UECPCommandDSNnPSN):
    ELEMENT_CODE = 0x3E

    def __init__(
        self, programme_type_name="", data_set_number=0, programme_service_number=0
    ):
        super().__init__(
            data_set_number=data_set_number,
            programme_service_number=programme_service_number,
        )
        self.__programme_type_name = ""
        self.programme_type_name = programme_type_name

    @property
    def programme_type_name(self) -> str:
        return self.__programme_type_name

    @programme_type_name.setter
    def programme_type_name(self, new_programme_type_name: str):
        new_programme_type_name = str(new_programme_type_name)
        if len(new_programme_type_name) > 8:
            raise InvalidProgrammeTypeName(
                new_programme_type_name, f"PTYN supports only 8 characters"
            )
        new_programme_type_name = new_programme_type_name.rstrip(" ")
        try:
            new_programme_type_name.encode("basic_rds_character_set")
        except ValueError as e:
            raise InvalidProgrammeServiceName(
                new_programme_type_name, f"PTYN cannot be encoded, exc={e!r}"
            )
        self.__programme_type_name = new_programme_type_name

    def encode(self) -> list[int]:
        return [
            self.ELEMENT_CODE,
            self.data_set_number,
            self.programme_service_number,
        ] + list(self.__programme_type_name.ljust(8).encode("basic_rds_character_set"))

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["ProgrammeTypeNameSetCommand", int]:
        data = list(data)
        if len(data) < 11:
            raise UECPCommandDecodeNotEnoughData(len(data), 11)

        mec, dsn, psn = data[0:3]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)

        programme_type_name = bytes(data[3:11]).decode("basic_rds_character_set")

        return (
            cls(
                programme_type_name=programme_type_name,
                data_set_number=dsn,
                programme_service_number=psn,
            ),
            11,
        )


@enum.unique
class RadioTextBufferConfiguration(enum.IntEnum):
    TRUNCATE_BEFORE = 0b00
    APPEND = 0b10


class InvalidNumberOfTransmissions(UECPCommandException):
    pass


def _ensure_radio_text_carriage_return(*values: str) -> str:
    if len(values) < 1:
        import inspect

        raise TypeError(
            f"{inspect.stack()[0][3]} must be called with at least one argument and the last one must be a string"
        )
    value = values[-1]
    value = str(value)
    if 0 < len(value) < 61 and value[-1] != "\r":
        value += "\r"
        warnings.warn(
            "Implicitly appending a carriage return 0x0D to radio text shorter than 61 characters"
        )
    return value


def _check_radio_text(_, __, value: str):
    value = str(value)
    if len(value) > 64 or len(value) == 0:
        raise ValueError(
            f"Radio text supports only up to 64 characters and must not be empty, given {value!r}",
        )
    if len(value) < 61 and value[-1] != "\r":
        raise ValueError(
            f"Radio text shorter than 61 characters must be terminated by a carriage return, given {value!r}"
        )
    value.encode("basic_rds_character_set")

    return value


@attr.s
class RadioText:
    text: str = attr.ib(
        converter=_ensure_radio_text_carriage_return,
        validator=_check_radio_text,
        on_setattr=[
            _ensure_radio_text_carriage_return,  # type:ignore
            _check_radio_text,
        ],
    )
    number_of_transmissions: int = attr.ib(default=0, converter=int)
    a_b_toggle: bool = attr.ib(default=False, converter=bool)

    @number_of_transmissions.validator
    def _check_number_of_transmissions(self, _, new_not: int):
        if not (0x0 <= new_not <= 0xF):
            raise InvalidNumberOfTransmissions(new_not)


@UECPCommand.register_type
class RadioTextSetCommand(UECPCommand, UECPCommandDSNnPSN):
    ELEMENT_CODE = 0x0A
    INFINITE_TRANSMISSIONS = 0

    def __init__(
        self,
        text="",
        number_of_transmissions=0,
        a_b_toggle=False,
        buffer_configuration=RadioTextBufferConfiguration.TRUNCATE_BEFORE,
        data_set_number=0,
        programme_service_number=0,
        radiotext: RadioText = None,
    ):
        super().__init__(
            data_set_number=data_set_number,
            programme_service_number=programme_service_number,
        )
        if radiotext is not None:
            self._radiotext = radiotext
        else:
            self._radiotext = RadioText(
                text=text,
                number_of_transmissions=number_of_transmissions,
                a_b_toggle=a_b_toggle,
            )
        self._buffer_configuration = RadioTextBufferConfiguration(buffer_configuration)

    @property
    def text(self) -> str:
        return self._radiotext.text

    @text.setter
    def text(self, new_text: str):
        self._radiotext.text = new_text

    @property
    def number_of_transmissions(self) -> int:
        return self._radiotext.number_of_transmissions

    @number_of_transmissions.setter
    def number_of_transmissions(self, new_not: int):
        self._radiotext.number_of_transmissions = new_not

    @property
    def a_b_toggle(self) -> bool:
        return self._radiotext.a_b_toggle

    @a_b_toggle.setter
    def a_b_toggle(self, toggle: bool):
        self._radiotext.a_b_toggle = toggle

    @property
    def buffer_configuration(self) -> RadioTextBufferConfiguration:
        return self._buffer_configuration

    @buffer_configuration.setter
    def buffer_configuration(self, buffer_conf: RadioTextBufferConfiguration):
        self._buffer_configuration = RadioTextBufferConfiguration(buffer_conf)

    @property
    def radiotext(self) -> RadioText:
        return self._radiotext

    def encode(self) -> list[int]:
        data = [self.ELEMENT_CODE, self.data_set_number, self.programme_service_number]
        if (
            len(self._radiotext.text) == 0
            and self._buffer_configuration
            is RadioTextBufferConfiguration.TRUNCATE_BEFORE
        ):
            data.append(0)
        else:
            mel = 1 + len(self._radiotext.text)
            flags = (
                self._buffer_configuration << 5
                | self._radiotext.number_of_transmissions << 1
                | self._radiotext.a_b_toggle
            )
            data += [mel, flags]
            data += list(self._radiotext.text.encode("basic_rds_character_set"))
        return data

    @classmethod
    def create_from(
        cls, data: typing.Union[bytes, list[int]]
    ) -> tuple["RadioTextSetCommand", int]:
        data = list(data)
        if len(data) < 4:
            raise UECPCommandDecodeNotEnoughData(len(data), 4)
        mec, dsn, psn, mel = data[0:4]
        if mec != cls.ELEMENT_CODE:
            raise UECPCommandDecodeElementCodeMismatchError(mec, cls.ELEMENT_CODE)
        if mel == 0:
            return cls(data_set_number=0, programme_service_number=0), 4
        data = data[4:]
        if len(data) < mel:
            raise UECPCommandDecodeNotEnoughData(len(data), mel)
        flags = data[0]
        buffer_configuration = (flags & 0b0110_0000) >> 5
        number_of_transmission = (flags & 0b0001_1110) >> 1
        a_b_toggle = flags & 0b0000_0001
        text_data = bytes(data[1:mel])
        text = text_data.decode("basic_rds_character_set")

        return (
            cls(
                text=text,
                buffer_configuration=buffer_configuration,
                number_of_transmissions=number_of_transmission,
                a_b_toggle=a_b_toggle,
                data_set_number=dsn,
                programme_service_number=psn,
            ),
            4 + mel,
        )


# TODO AF
# TODO EON-AF
# TODO Slow Labeling Codes
# TODO Linkage information



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uecp', 'uecp.commands', 'uecp.serial_con']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.2.0,<22.0.0',
 'crc>=1.1.2,<2.0.0',
 'pyserial-asyncio>=0.6,<0.7',
 'pyserial>=3.5,<4.0']

setup_kwargs = {
    'name': 'uecp',
    'version': '0.3.0',
    'description': 'UECP encoder & decoder',
    'long_description': '# python UECP implementation\n\nThis library aims to support UECP version 6.02. It will start with basic commands required to support our needs.\n',
    'author': 'Christian Kohlstedde',
    'author_email': 'christian@kohlsted.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import codecs
import typing

from crc import Configuration, CrcCalculator  # type: ignore

from uecp.byte_stuffing_codec import (
    IncrementalDecoder as ByteStuffingIncrementalDecoder,
)
from uecp.commands.base import UECPCommand


class UECPFrame:
    STA = 0xFE
    STP = 0xFF

    ALL_SITES = 0
    ALL_ENCODERS = 0

    UNUSED_SEQUENCE_COUNTER = 0

    CRC_CONFIGURATION = Configuration(
        width=16,
        polynomial=0x1021,
        init_value=0xFFFF,
        final_xor_value=0xFFFF,
        reverse_input=False,
        reverse_output=False,
    )

    def __init__(
        self,
        site_address=0,
        encoder_address=0,
        sequence_counter=0,
        commands: typing.Optional[list[UECPCommand]] = None,
    ):
        self._site_address: int = 0
        self._encoder_address: int = 0
        self._sequence_counter: int = 0
        self._commands: list[UECPCommand] = []

        self.site_address = site_address
        self.encoder_address = encoder_address
        self.sequence_counter = sequence_counter

        self._assumed_command_length = 0
        if commands is not None:
            for command in commands:
                self.add_command(command)

    @property
    def site_address(self) -> int:
        return self._site_address

    @site_address.setter
    def site_address(self, new_site_address: int):
        if not (0 <= new_site_address <= 0x3FF):
            raise ValueError(
                f"Site address must be in range of 0 to 0x3ff, {new_site_address:#x} given"
            )
        if new_site_address != int(new_site_address):
            raise ValueError("Site address must be an integer")
        self._site_address = int(new_site_address)

    @property
    def encoder_address(self) -> int:
        return self._encoder_address

    @encoder_address.setter
    def encoder_address(self, new_encoder_address: int):
        if not (0 <= new_encoder_address <= 0x3F):
            raise ValueError(
                f"Encoder address must be in range of 0 to 0x3f, {new_encoder_address:#x} given"
            )
        if new_encoder_address != int(new_encoder_address):
            raise ValueError("Encoder address must be an integer")
        self._encoder_address = int(new_encoder_address)

    @property
    def sequence_counter(self) -> int:
        return self._sequence_counter

    @sequence_counter.setter
    def sequence_counter(self, new_sequence_counter: int):
        if not (0 <= new_sequence_counter <= 0xFF):
            raise ValueError(
                f"Sequence counter must be in range of 0 to 0xff, {new_sequence_counter:#x} given"
            )
        if new_sequence_counter != int(new_sequence_counter):
            raise ValueError("Sequence counter must be an integer")
        self._sequence_counter = int(new_sequence_counter)

    def add_command(self, *commands: UECPCommand):
        for command in commands:
            cmd_len = len(command.encode())
            if self._assumed_command_length + cmd_len > 255:
                raise OverflowError()
            self._assumed_command_length += cmd_len
            self._commands.append(command)

    def clear_commands(self):
        self._commands = []

    @property
    def commands(self) -> list[UECPCommand]:
        return list(self._commands)

    def encode(self) -> bytes:
        data: list[int] = []

        # address composed by 10 bits sites address & 6 bits encoder address
        address = self._site_address << 6 | self._encoder_address

        data.append(address >> 8)
        data.append(address & 0xFF)
        data.append(self._sequence_counter)

        msg_data: list[int] = []
        for command in self._commands:
            msg_data += command.encode()

        if not (0 <= len(msg_data) <= 255):
            raise ValueError("Encoded commands must not exceed 255 bytes")

        data.append(len(msg_data))
        data += msg_data

        crc_calculator = CrcCalculator(self.CRC_CONFIGURATION)
        crc = crc_calculator.calculate_checksum(data)
        data.append(crc >> 8)
        data.append(crc & 0xFF)

        data = list(codecs.encode(data, "uecp_frame"))  # type: ignore

        return bytes([self.STA] + data + [self.STP])

    @classmethod
    def create_from_enclosed(cls, data: typing.Union[bytes, list[int]]) -> "UECPFrame":
        if len(data) < 6:
            raise ValueError("not enough data")

        data, crc_high, crc_low = data[:-2], data[-2], data[-1]
        crc = crc_high << 8 | crc_low

        crc_calculator = CrcCalculator(cls.CRC_CONFIGURATION)
        crc_computed = crc_calculator.calculate_checksum(data)
        if crc != crc_computed:
            raise ValueError(f"CRC error {crc} vs {crc_computed}")

        address_high, address_low, sequence_counter = data[0:3]
        msg_len, msg_data = data[3], data[4:]
        if msg_len != len(msg_data):
            raise ValueError(
                f"Data length doesn't match, expected {msg_len}, given {len(msg_data)}"
            )

        address = address_high << 8 | address_low

        site_address = address >> 6
        encoder_address = address & 0x3F

        commands = UECPCommand.decode_commands(msg_data)

        return cls(
            site_address=site_address,
            encoder_address=encoder_address,
            sequence_counter=sequence_counter,
            commands=commands,
        )


class UECPFrameDecoder:
    def __init__(self):
        self._start_bit_seen = False
        self._enclosed_data = []
        self._enclosed_incremental_decoder = ByteStuffingIncrementalDecoder()

    def decode(
        self, data: typing.Union[bytes, list[int]]
    ) -> tuple[typing.Optional[UECPFrame], typing.Union[bytes, list[int]]]:
        i = 0

        try:
            for i, byte in enumerate(data, start=1):
                if byte == UECPFrame.STA:
                    self._start_bit_seen = True
                elif byte == UECPFrame.STP:
                    if self._start_bit_seen is False:
                        raise ValueError("Stop bit seen, but no start bit")
                    if len(self._enclosed_data) <= 1:
                        raise ValueError("No payload data decoded")
                    frame = UECPFrame.create_from_enclosed(self._enclosed_data)
                    self.reset()
                    return frame, data[i:]
                else:
                    self._enclosed_data += list(
                        self._enclosed_incremental_decoder.decode([byte])
                    )
        except Exception as e:
            self.reset()
            raise e

        return None, data[i:]

    def reset(self):
        self._enclosed_data.clear()
        self._enclosed_incremental_decoder.reset()
        self._start_bit_seen = False

    @property
    def empty(self) -> bool:
        return (
            len(self._enclosed_data) == 0
            and self._enclosed_incremental_decoder.getstate()[1] == 1
        )
